
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Author: rostami@google.com (Afshin Rostamizadeh)

#include <iostream>

#include "kernel/lib/kernel-learning-linearl1.h"
#include "kernel/lib/kernel-learning-linearl2.h"
#include "kernel/lib/kernel-learning-polyl2.h"
#include "kernel/lib/kernel-learning-alignment.h"
#include "kernel/lib/io-utilities.h"

using namespace kernel;

DEFINE_double(alg_reg, 1.0,
              "Regularization parameter used for learning algorithm.");
DEFINE_double(ker_reg, 1.0,
              "Regularization parameter used for the kernel selection.");
DEFINE_double(tol, 1e-5, "Stopping criteria for iterative methods.");
DEFINE_int32(max_iter, 50,
             "Maximum number of iterations used for iterative methods.");
DEFINE_string(mu_file, "", "Save weights to output file.");
DEFINE_string(lk_alg, "unif",
     "\tunif -- uniform weights (i.e. standard linear kernel)\n"
     "\tquadl2 -- l2-regularized quadratic combination\n"
     "\tlin1 -- l1-regularized linear combination (classification only).\n"
     "\tlin2 -- l2-regularized linear combination\n"
     "\talign -- weights proportional to alignment\n"
     "\talignf -- (positive) weights which maximize final kernel alignment");
DEFINE_string(num_train, "0:0",
        "Use the points in the range N:M (i.e. N, N+1, ..., M for training.\n"
        "0:0 indicated using all the points");
DEFINE_double(mu0, 0.0,
              "Set (uniform) value of mu0 vector, for use with lin2 method");
DEFINE_double(mu0_align, 0,
              "Set value of mu0 vector to align weights, with specified norm");
DEFINE_double(interp_param, 0.5,
              "Controls step size for gradient descent method used with\n"
              "lin2 method");
DEFINE_string(pred_file, "",
          "Output predictions to pred_file, if possible.");
DEFINE_bool(read_kar, false,
          "Read kernels as OpenKernel kar objects.");
DEFINE_string(label_file, "", 
          "File containing one label per line (required when using kar flag)");
DEFINE_bool(trace_normalize, false,
            "Normalize each base kernel: K' = K / Trace(K).");
DEFINE_bool(frob_normalize, false,
            "Normalize each base kernel: K' = K / Frobenius(K).");
DEFINE_bool(align_normalize, false,
            "Normalize each base kernel: K' = K * align(K).");
DEFINE_bool(alignf_normalize, false,
            "Normalize each base kernel according to alignf weights.");
DEFINE_bool(center_labels, false,
            "Center the training labels, y' = y - mean(y).");
DEFINE_bool(center_kernels, false,
        "Center the data in feature space, phi(x)' = phi(x) - mean(phi(x)).");
DEFINE_double(alignos_reg, 1.0,
       "Additional regularization parameter used for the alignos algorithm.");

int main(int argc, char** argv) {
  string usage =
      "Output a combination of input kernels"
      "according to the selected automatic kernel selection method.\n\n";
  usage += "Usage: " + string(argv[0]);
  usage += " kernel.list outfile\n";

  std::set_new_handler(FailedNewHandler);
  SetFlags(usage.c_str(), &argc, &argv, true);

  if (argc < 3) {
    ShowUsage();
    return 1;
  }

  // Parse filename list
  vector<string> filenames;

  string line;
  ifstream file(argv[1]);
  if(!file.is_open()) {
    LOG(ERROR) << "Error: Could not read kernel filename list: "
         << argv[1] << "." << endl;
    return 1;
  }
  while(1) {
    getline(file, line);
    if(file.eof()) break;
    filenames.push_back(line);
  }
  file.close();

  // Load data
  DenseVector<double> labels, pids;
  vector< DenseMatrix<double> > data_vec;
  int flag;

  if (FLAGS_read_kar) {
    if(FLAGS_label_file.compare("") == 0) 
      LOG(FATAL) << "Label file is required when using kar format.";
    AsciiToDenseVector(FLAGS_label_file, &labels);

    for (size_t i = 0; i < filenames.size(); ++i) {
      DenseMatrix<double> data;
      KarToDense(filenames[i], &data);
      data_vec.push_back(data);
   }

  } else { //read libsvm kernel format

    for (size_t i = 0; i < filenames.size(); ++i) {
      DenseMatrix<double> data;
      flag = LibSVMKernelToDense(filenames[i], &data, &labels, &pids);
      if(flag != 0) {
        LOG(ERROR) << "Error reading input file: " << filenames[i]
                   << "." << endl;
        LOG(ERROR) << "Code: " << flag << endl;
        return -1;
      }
      data_vec.push_back(data);
    }
  }

  // Select training points
  vector< DenseMatrix<double> > data_train_vec;
  DenseVector<double> labels_train;

  // parse num_train flag
  vector<char*> tokens;
  size_t train_start, train_end;
  char *cstr = new char[FLAGS_num_train.size()+1];
  strcpy(cstr, FLAGS_num_train.c_str());
  SplitToVector(cstr, ":", &tokens, true);
  if (tokens.size() != 2)
    LOG(FATAL) << "Format error in num_train flag";
  train_start = atoi(tokens[0]);
  train_end = atoi(tokens[1]);
  delete[] cstr;

  if (train_start < train_end) {
    DenseVector<size_t> train_points;
    //MakeRange(train_start, train_end, &train_points);
    MakeRangeWrap(train_start, train_end, data_vec[0].NumRow(), &train_points);

    // Need to center before spliting data
    if (FLAGS_center_kernels)
      for (size_t i = 0; i < data_vec.size(); ++i)
        CenterKernel(train_points, &data_vec[i]);

    for (size_t i = 0; i < data_vec.size(); ++i) {
      DenseMatrix<double> temp;
      SubMatrixCopy(data_vec[i], train_points, train_points, &temp);
      data_train_vec.push_back(temp);
    }

    SubsetCopy(labels, train_points, &labels_train);
  } else {
    if (FLAGS_center_kernels) 
      for (size_t i = 0; i < data_vec.size(); ++i)
        CenterKernel(&data_vec[i]);
   
    data_train_vec = data_vec; //TODO(rostami): should not have to copy data
    labels_train = labels;
  }

  if (FLAGS_center_labels) {
    Sum(-1 * Mean(labels_train), DenseVector<double>(labels.Size(), 1.0),
        &labels);
    Sum(-1 * Mean(labels_train), DenseVector<double>(labels_train.Size(), 1.0),
        &labels_train);
  }

  if (FLAGS_trace_normalize) {
    for (size_t i = 0; i < data_train_vec.size(); ++i) {
      ScalarProduct(1 / Trace(data_train_vec[i]), &data_vec[i]);
      ScalarProduct(1 / Trace(data_train_vec[i]), &data_train_vec[i]);
    }
  }

  if (FLAGS_frob_normalize) {
    for (size_t i = 0; i < data_train_vec.size(); ++i) {
      ScalarProduct(1 / FrobeniusNorm(data_train_vec[i]), &data_vec[i]);
      ScalarProduct(1 / FrobeniusNorm(data_train_vec[i]), &data_train_vec[i]);
    }
  }


  if (FLAGS_align_normalize) {
    DenseVector<double> alignments(data_train_vec.size(), 0);
    MakeAlignmentWeights(data_train_vec, labels_train, data_train_vec.size(),
      !FLAGS_center_kernels, !FLAGS_center_labels, &alignments);

    for (size_t i = 0; i < data_train_vec.size(); ++i) {
      ScalarProduct(alignments.Element(i), &data_vec[i]);
      ScalarProduct(alignments.Element(i), &data_train_vec[i]);
    }
  }

  if (FLAGS_alignf_normalize) {
    DenseVector<double> alignments(data_train_vec.size(), 0);
    MakeFullAlignmentWeights(data_train_vec, labels_train, 
      data_train_vec.size(), !FLAGS_center_kernels, !FLAGS_center_labels,
      &alignments);

    for (size_t i = 0; i < data_train_vec.size(); ++i) {
      if (alignments.Element(i) > 1e-9) {
        ScalarProduct(alignments.Element(i), &data_vec[i]);
        ScalarProduct(alignments.Element(i), &data_train_vec[i]);
      } else {
        data_vec.erase(data_vec.begin() + i);
        data_train_vec.erase(data_train_vec.begin() + i);
      }
    }
  }

  // Learn combination weights
  DenseVector<double> weights;
  DenseVector<double> alg_weights;
  double offset = 0;
  if (FLAGS_lk_alg.compare("unif") == 0) {
    DenseVector<double> uniform(data_train_vec.size(),
                          1.0 / data_train_vec.size() * FLAGS_ker_reg);
    weights = uniform;

  } else if (FLAGS_lk_alg.compare("lin2") == 0) {
    // setup mu0 parameter
    DenseVector<double> mu0(data_train_vec.size(), FLAGS_mu0);
    if (FLAGS_mu0_align > 0)
      MakeFullAlignmentWeights(data_train_vec, labels_train, FLAGS_mu0_align,
        !FLAGS_center_kernels, !FLAGS_center_labels, &mu0);

    MakeLinearL2Weights(data_train_vec, labels_train,
                       mu0, FLAGS_ker_reg, FLAGS_alg_reg, FLAGS_interp_param,
                       FLAGS_tol, FLAGS_max_iter, &weights, &alg_weights);

  } else if (FLAGS_lk_alg.compare("lin1") == 0) {
    MakeLinearL1Weights(data_train_vec, false, labels_train, FLAGS_alg_reg,
                        FLAGS_tol, FLAGS_max_iter, &weights, &alg_weights,
                        &offset);

  } else if (FLAGS_lk_alg.compare("quadl2") == 0) {
    DenseMatrix<double> empty_data;
    MakePolyL2Weights(data_train_vec, empty_data, labels_train,
                      FLAGS_ker_reg, FLAGS_alg_reg, FLAGS_interp_param,
                      FLAGS_tol, FLAGS_max_iter, &weights, &alg_weights);

  } else if (FLAGS_lk_alg.compare("align") == 0) {
    // if labels and kernels are already centered, do not need to center again
    MakeAlignmentWeights(data_train_vec, labels_train, FLAGS_ker_reg,
      !FLAGS_center_kernels, !FLAGS_center_labels,
      &weights);

  } else if (FLAGS_lk_alg.compare("alignf") == 0) {
    // if labels and kernels are already centered, do not need to center again
    MakeFullAlignmentWeights(data_train_vec, labels_train, FLAGS_ker_reg,
      !FLAGS_center_kernels, !FLAGS_center_labels, &weights);

  } else if (FLAGS_lk_alg.compare("alignos") == 0) {
    // if labels and kernels are already centered, do not need to center again
    MakeOnestageAlignmentWeights(data_train_vec, labels_train, 1, 
      FLAGS_alg_reg, FLAGS_ker_reg, FLAGS_alignos_reg,
      !FLAGS_center_kernels, !FLAGS_center_labels, FLAGS_max_iter,
      &weights);

  } else {
    LOG(ERROR) << "Unknown weight type" << endl;
    return 1;
  }

  int nnz = 0;
  for (size_t i = 0; i < weights.Size(); ++i) {
    if (abs(weights.Element(i)) > 1e-9)
      ++nnz;
  }
  LOG(INFO) << "Number of non-zero weights: " << nnz;

  // Save weights if requested
  if (FLAGS_mu_file.compare("") != 0) {
    ofstream out(FLAGS_mu_file.c_str());
    if(!out.is_open()) {
      LOG(ERROR) << "Unable to save weights" << endl;
    } else {
      PrintVector(weights, out);
      out.close();
    }
  }

  // Combine kernels and print output
  DenseMatrix<double> kernel;
  LinearCombination(data_vec, weights, &kernel);
  if (FLAGS_lk_alg.compare("quadl2") == 0) {
    PointwisePower(2, &kernel);
  }
  ofstream out(argv[2]);
  if(!out.is_open()) {
    LOG(ERROR) << "Unable to output kernel" << endl;
  } else {
    KernelPrintLibSVM(kernel, labels, pids, out);
    out.close();
  }

  // Save predictions if requested and if availible
  if (FLAGS_pred_file.compare("") != 0 && alg_weights.Size() > 0) {

    DenseMatrix<double> k_test;
    if (train_start < train_end) {
      DenseVector<size_t> train_points;
      MakeRangeWrap(train_start, train_end,
                    data_vec[0].NumRow(), &train_points);
      SubMatrixCopy(kernel, train_points, &k_test);
      k_test.Transpose(&k_test);
    } else {
      k_test = kernel;
    }

    ofstream out(FLAGS_pred_file.c_str());
    if(!out.is_open()) {
      LOG(ERROR) << "Unable to save alg_weights" << endl;
    } else {

      DenseVector<double> pred;
      if (FLAGS_lk_alg.compare("lin1") == 0) { //svm pred
        SVMDualPredict(k_test, alg_weights, labels, offset, &pred);
      } else { //krr pred
        KRRDualPredict(k_test, alg_weights, &pred);
      }

      PrintVector(pred, out);
    }
  }

  return 0;
}
