// dense-matrix.h
//

// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Author: allauzen@google.com (Cyril Allauzen)
//
// \file
// A simple reference-counted dense matrix representation

#ifndef KERNEL_LIB_DENSE_MATRIX_H_
#define KERNEL_LIB_DENSE_MATRIX_H_

#include "kernel/lib/blas.h"
#include "kernel/lib/sparse-vector.h"
#include "kernel/lib/kernel.h" // change to min req


namespace kernel {

//
// DENSE MATRIX CLASS
//

// Forward declarations
template <typename T>
class DenseMatrixData;
template <class T>
class DenseVector;

// DenseMatrix: a simple reference-counted dense matrix representation
template <typename T>
class DenseMatrix {
 public:
  typedef T ElementType;
  typedef size_t IndexType;
  typedef DenseMatrix<T> TransposeMatrix;

  DenseMatrix() : data_(new DenseMatrixData<T>), transpose_(false) {}

  DenseMatrix(size_t num_row, size_t num_col, const T &x = T())
      : data_(new DenseMatrixData<T>(num_row, num_col, x)),
        transpose_(false) {}

  DenseMatrix(const DenseMatrix<T> &m)
  : data_(m.data_), transpose_(m.transpose_) {
    if (data_)
      data_->IncrRefCount();
  }
  explicit DenseMatrix(const DenseVector<T> &v)
      : data_(new DenseMatrixData<T>(v.Size(), 1, T())), transpose_(false) {
    for (size_t i = 0; i < v.Size(); ++i)
      data_->SetElement(i, 0, v.Element(i));
  }

  ~DenseMatrix() {
    if (!data_->DecrRefCount())
      delete data_;
  }

  DenseMatrix<T> &operator=(const DenseMatrix<T> &m) {
    if (this != &m) {
      if (!data_->DecrRefCount()) delete data_;
      m.data_->IncrRefCount();
      data_ = m.data_;
      transpose_ = m.transpose_;
    }
    return *this;
  }

  size_t NumRow() const {
    return transpose_ ? data_->NumColumn() : data_->NumRow();
  }

  size_t NumColumn() const {
    return transpose_ ? data_->NumRow() : data_->NumColumn();
  }

  void Reset(size_t num_row, size_t num_col, const T &x = T()) {
    MutateCheck();
    data_->Reset(num_row, num_col, x);
    transpose_ = false;
  }

  void Clear() {
    MutateCheck();
    data_->Reset(0, 0, 0);
    transpose_ = false;
  }


  const T& Element(size_t i, size_t j) const {
    return transpose_ ? data_->Element(j, i) : data_->Element(i, j);
  }

  void SetElement(size_t i, size_t j, const T &x) {
    MutateCheck();
    CHECK(i < this->NumRow());
    CHECK(j < this->NumColumn());
    if (transpose_)
      data_->SetElement(j, i, x);
    else
      data_->SetElement(i, j, x);
  }

  void Transpose(TransposeMatrix  *m) const {
    *m = TransposeMatrix(data_, !transpose_);
  }

  bool IsTranspose() {
    return transpose_;
  }

  void SetBlasInfo(BlasMatrixInfo<T> *info) const {
    info->transpose = transpose_ ? 't' : 'n';
    info->leading_dimension = data_->NumRow();
    info->other_dimension = data_->NumColumn();
    info->array = data_->CArray();
  }

  void SetBlasInfo(MutableBlasMatrixInfo<T> *info) {
    MutateCheck();
    info->transpose = transpose_ ? 't' : 'n';
    info->leading_dimension = data_->NumRow();
    info->other_dimension = data_->NumColumn();
    info->array = data_->CArray();
  }

 private:
  DenseMatrix(DenseMatrixData<T> *data, bool transpose)
      : data_(data), transpose_(transpose) {
    data_->IncrRefCount();
  }

  void MutateCheck() {
    // Copy on write
    if (data_->RefCount() > 1) {
      data_->DecrRefCount();
      data_ = new DenseMatrixData<T>(*data_);
    }
  }

  DenseMatrixData<T> *data_;
  bool transpose_;
};


// Provides the data representation of a dense matrix as
// a STL vector in column-major ordering.
template <typename T>
class DenseMatrixData {
 public:
  DenseMatrixData() : num_row_(0), num_col_(0) {}

  DenseMatrixData(size_t num_row, size_t num_col, const T &x)
      : elements_(num_row * num_col, x), num_row_(num_row),
        num_col_(num_col) {}

  explicit DenseMatrixData(const DenseMatrixData<T> &data)
      : elements_(data.elements_), num_row_(data.num_row_),
        num_col_(data.num_col_) {}

  ~DenseMatrixData() {}

  size_t Size() const { return elements_.size(); }
  size_t NumRow() const { return num_row_; }
  size_t NumColumn() const { return num_col_; }

  void Reset(size_t num_row, size_t num_col, const T &x) {
    elements_.clear();
    elements_.resize(num_row * num_col, x);
    num_row_ = num_row;
    num_col_ = num_col;
  }

  const T &Element(size_t i, size_t j) const {
    return elements_[i + num_row_ * j];
  }

  void SetElement(size_t i, size_t j, const T &x) {
    elements_[i + num_row_ * j] = x;
  }

  int RefCount() const { return ref_count_.count(); }
  int IncrRefCount() { return ref_count_.Incr(); }
  int DecrRefCount() { return ref_count_.Decr(); }

  const T *CArray() const { return &(elements_[0]); }

  T *CArray() { return const_cast<T*>(&(elements_[0])); }

 private:
  vector<T> elements_;
  size_t num_row_;
  size_t num_col_;
  RefCounter ref_count_;

  void operator=(const DenseMatrix<T> &);  // disallow
};


//
// MATRIX OPERATIONS
//

// Copy the submatrix defined by the indices row_index, column_index.
// Result matrix:
//      submatrix(i,j) = mat(row_index(i), column_index(j))
template <typename T, typename I /*=int*/>
void SubMatrixCopy(const DenseMatrix<T> &mat,
                   const DenseVector<I> &row_index,
                   const DenseVector<I> &column_index,
                   DenseMatrix<T> *submatrix) {
  if (submatrix->NumRow() != row_index.Size() ||
     submatrix->NumColumn() != column_index.Size())
    submatrix->Reset(row_index.Size(), column_index.Size(), 0);

  int row, col;
  for (size_t i = 0; i < row_index.Size(); ++i) {
    row = row_index.Element(i);
    CHECK(row < mat.NumRow());
    for (size_t j = 0; j < column_index.Size(); ++j) {
      col = column_index.Element(j);
      CHECK(col < mat.NumColumn());

      submatrix->SetElement(i, j, mat.Element(row, col));
    }
  }
}

template <typename T, typename I /*=int*/>
void SubMatrixCopy(const DenseMatrix<T> &mat,
                   const DenseVector<I> &row_index,
                   DenseMatrix<T> *submatrix) {

  DenseVector<I> all_col;
  MakeRange(0, mat.NumColumn(), &all_col);
  SubMatrixCopy(mat, row_index, all_col, submatrix);
}

// Return a vector of sparse vectors representing the columns of mat
template <typename T>
void MatrixToColumnVectors(const DenseMatrix<T> &mat,
                          vector< SparseVector<T> > *columns) {
  columns->clear();
  for (size_t i = 0; i < mat.NumColumn(); ++i) {
    SparseVector<T> column;
    for (size_t j = 0; j <mat.NumRow(); ++j)
      column.AddElement(j, mat.Element(j, i));
    columns->push_back(column);
  }
}

// Matrix Trace: sum of diagonal entries
template <typename T>
T Trace(const DenseMatrix<T> &m) {
  T sum = 0;
  size_t min_dim = (m.NumRow() < m.NumColumn()) ? m.NumRow() : m.NumColumn();
  for (size_t i = 0; i < min_dim; ++i)
    sum += m.Element(i, i);

  return sum;
}

// Frobenius Norm (squared): sum of squares of all entries
template <typename T>
T FrobeniusNorm(const DenseMatrix<T> &m) {
  T sum = 0;
  for (size_t i = 0; i < m.NumRow(); ++i)
    for (size_t j = 0; j < m.NumColumn(); ++j)
      sum += pow(m.Element(i,j), 2);

  return sum;
}

// Frobeinus Product
template <typename T> 
T FrobeniusProductSym(const DenseMatrix<T> &a, const DenseMatrix<T> &b) {
  CHECK(a.NumRow() == a.NumColumn());
  CHECK(b.NumRow() == b.NumColumn());
  CHECK(a.NumRow() == b.NumRow());

  T sum = 0;
  for (size_t i = 0; i < a.NumRow(); ++i)
    for (size_t j = 0; j < i; ++j)
      sum += a.Element(i,j) * b.Element(i,j) * 2;

  for (size_t i = 0; i < a.NumRow(); ++i)
    sum += a.Element(i,i) * b.Element(i,i);

  return sum;
}

// Given a DenseVector, output a diagonal DenseMatrix with diagonal equal
// to the input vector.
template<typename T> inline
void Diagonal(const DenseVector<T> &diag_vector,
              DenseMatrix<T> *I) {
  I->Reset(diag_vector.Size(), diag_vector.Size(), 0);
  for (size_t i = 0; i < diag_vector.Size(); ++i)
    I->SetElement(i, i, diag_vector.Element(i));
}

// Matrix equality:  A == B ?
template <typename T>
bool Equal(const DenseMatrix<T> &a, const DenseMatrix<T> &b) {
  if (a.NumRow() != b.NumRow()) return false;
  if (a.NumColumn() != b.NumColumn()) return false;

  for (size_t i = 0; i < a.NumRow(); ++i)
    for (size_t j = 0; j < a.NumColumn(); ++j)
      if (a.Element(i, j) != b.Element(i, j))
        return false;

  return true;
}

// Matrix approximate equality:  A == B +/- eps?
template <typename T>
bool DenseMatrixApproxEqual(const DenseMatrix<T> &a, const DenseMatrix<T> &b,
                 T eps) {
  if (a.NumRow() != b.NumRow()) return false;
  if (a.NumColumn() != b.NumColumn()) return false;

  for (size_t i = 0; i < a.NumRow(); ++i) {
    for (size_t j = 0; j < a.NumColumn(); ++j) {
      if (a.Element(i, j) > b.Element(i, j) + eps)
        return false;
      if (a.Element(i, j) < b.Element(i, j) - eps)
        return false;
		}
	}

  return true;
}

// Matrix approximate equality
template <typename T>
bool ApproxEqual(const DenseMatrix<T> &a,
                 const DenseMatrix<T> &b,
                 T tol) {
  if (a.NumRow() != b.NumRow()) return false;
  if (a.NumColumn() != b.NumColumn()) return false;

  for (size_t i = 0; i < a.NumRow(); ++i)
    for (size_t j = 0; j < a.NumColumn(); ++j)
      if (   a.Element(i, j) > b.Element(i, j) + tol
          || a.Element(i, j) < b.Element(i, j) - tol) {
        VLOG(2) << "element (" << i << "," << j << ") differ: "
             << a.Element(i, j) << ", " << b.Element(i, j) << endl;
        return false;
      }

  return true;
}

// Maximum entry of the matrix
template <typename T>
T Max(const DenseMatrix<T> &a) {

  T max = a.Element(0,0);

  for (size_t i = 0; i < a.NumRow(); ++i) {
    for (size_t j = 0; j < a.NumColumn(); ++j) {
      if (max < a.Element(i,j))
        max = a.Element(i,j);
    }
  }

  return max;
}

//
// Matrix addition
//
// C = alpha * A + beta * B
template <typename T>
void Sum(const T &alpha, const DenseMatrix<T> &a,
         const T &beta, const DenseMatrix<T> &b,
         DenseMatrix<T> *c) {
  CHECK(a.NumRow() == b.NumRow());
  CHECK(a.NumColumn() == b.NumColumn());
  if((c->NumRow() != a.NumRow()) ||  (c->NumColumn() != a.NumColumn()))
    c->Reset(a.NumRow(), a.NumColumn(), 0);

  for (size_t i = 0; i < a.NumRow(); ++i)
    for (size_t j = 0; j < a.NumColumn(); ++j)
      c->SetElement(i, j, alpha * a.Element(i, j) + beta * b.Element(i, j));
}

// C = alpha * A + beta * C
template <typename T> inline
void Sum(const T &alpha, const DenseMatrix<T> &a,
         const T &beta, DenseMatrix<T> *c) {
  Sum(alpha, a, beta, *c, c);
}

// C = A + B
template <typename T> inline
void Sum(const DenseMatrix<T> &a, const DenseMatrix<T> &b,
         DenseMatrix<T> *c) {
  return Sum<T>(1.0F, a, 1.0F, b, c);
}

// C = A + C
template <typename T> inline
void Sum(const DenseMatrix<T> &a, DenseMatrix<T> *c) {
  return Sum<T>(1.0F, a, 1.0F, c);
}

// Sum of all entries of the matrix
template <typename T> inline
T Sum(const DenseMatrix<T> &a) {
  DenseVector<T> ones_row(a.NumRow(), 1.0);
  DenseVector<T> ones_col(a.NumColumn(), 1.0);
  DenseVector<T> temp(a.NumRow(), 0.0);
  Product(a, ones_col, &temp);
  return DotProduct(ones_row, temp);
}

// Sum of the indicated row of the matrix
template <typename T> inline
T Sum(const DenseMatrix<T> &a, size_t row) {
  double sum = 0;
  for (size_t i = 0; i < a.NumColumn(); ++i)
    sum += a.Element(row, i); 

  return sum;
}

// Sum the indicated row and columns of the matrix
template <typename T, typename U> inline
T Sum(const DenseMatrix<T> &a,
      const DenseVector<U> &rows, const DenseVector<U> &columns) {
  double sum = 0;
  for (size_t i = 0; i < rows.Size(); ++i)
    for (size_t j = 0; j < columns.Size(); ++j)
      sum += a.Element(rows.Element(i), columns.Element(j));

  return sum;
}

//
// Matrix multiplication:
//
// C = alpha * A * B + beta * C
template <typename T>
void MatrixMatrixProduct(T alpha,
             const DenseMatrix<T> &a,
             const DenseMatrix<T> &b,
             T beta,
             DenseMatrix<T> *c) {
  CHECK(a.NumColumn() == b.NumRow());
  if ((a.NumRow() != c->NumRow() || b.NumColumn() == c->NumColumn()))
    c->Reset(a.NumRow(), b.NumColumn(), static_cast<T>(0.0F));
  BlasMatrixInfo<T> info_a, info_b;
  MutableBlasMatrixInfo<T> info_c;
  a.SetBlasInfo(&info_a);
  b.SetBlasInfo(&info_b);
  c->SetBlasInfo(&info_c);
  CHECK(info_c.transpose == 'n');
  int32 m = a.NumRow();
  int32 n = b.NumColumn();
  int32 k = a.NumColumn();

  Gemm<T>()(&(info_a.transpose),
            &(info_b.transpose),
            &m,
            &n,
            &k,
            &alpha,
            info_a.array,
            &(info_a.leading_dimension),
            info_b.array,
            &(info_b.leading_dimension),
            &beta,
            info_c.array,
            &(info_c.leading_dimension));
}

// C = A * B
template <typename T> inline
void Product(const DenseMatrix<T> &a,
             const DenseMatrix<T> &b,
             DenseMatrix<T> *c) {

  if ((a.NumRow() != c->NumRow() || b.NumColumn() == c->NumColumn()))
    c->Reset(a.NumRow(), b.NumColumn(), static_cast<T>(0.0F));
  MatrixMatrixProduct<T>(static_cast<T>(1.0F), a, b, static_cast<T>(0.0F), c);
  //Product(static_cast<T>(1.0F), a, b, static_cast<T>(0.0F), c);
}

// B = A * B
// In place product, slower but useful if B is large and A is small.
// A must be square.
// TODO(rostami): use blas to compute chunks of the product, this could
// be the best time/space tradeoff.
template <typename T> inline
void Product(const DenseMatrix<T> &a,
             DenseMatrix<T> *b) {
  CHECK(a.NumColumn() == b->NumRow());
  CHECK(a.NumColumn() == a.NumRow());

  DenseVector<T> scratch(b->NumRow(), 0);
  T temp;
  for (size_t i = 0; i < b->NumColumn(); ++i) {
    for (size_t j = 0; j < b->NumRow(); ++j)
      scratch.SetElement(j, b->Element(j, i));

    for (size_t j = 0; j < a.NumRow(); ++j) {
      temp = 0;
      for (size_t k = 0; k < a.NumColumn(); ++k)
        temp += a.Element(j, k) * scratch.Element(k);
      b->SetElement(j, i, temp);
    }
 }
}

// Point-wise matrix product.
// C = A .* B
template <typename T> inline
void PointwiseProduct(const DenseMatrix<T> &a, const DenseMatrix<T> &b,
                      DenseMatrix<T> *c) {
  CHECK(a.NumRow() == b.NumRow());
  CHECK(a.NumColumn() == b.NumColumn());
  if ((c->NumRow() != a.NumRow()) || (c->NumColumn() != a.NumColumn()))
    c->Reset(a.NumRow(), a.NumColumn(), 0);
  for (size_t i = 0; i < a.NumRow(); ++i)
    for (size_t j = 0; j < a.NumColumn(); ++j)
      c->SetElement(i, j, a.Element(i, j) * b.Element(i, j));
}

// Point-wise (in-place) matrix product.
// B = A .* B
template <typename T> inline
void PointwiseProduct(const DenseMatrix<T> &a, DenseMatrix<T> *b) {
  PointwiseProduct(a, *b, b);
}

//
// Multiplication by a scalar
//
// C = alpha * A
template <typename T>
void ScalarProduct(const T &alpha,
                   const DenseMatrix<T> &a,
                   DenseMatrix<T> *c) {
  if ((c->NumRow() != a.NumRow()) || (c->NumColumn() != a.NumColumn()))
    c->Reset(a.NumRow(), a.NumColumn(), 0);

  for (size_t i = 0; i < a.NumRow(); ++i)
    for (size_t j = 0; j < a.NumColumn(); ++j)
      c->SetElement(i, j, alpha * a.Element(i,j));
}

// C = alpha * C
template <typename T> inline
void ScalarProduct(const T &alpha, DenseMatrix<T> *c) {
  ScalarProduct(alpha, *c, c);
}


//
// MATRIX-VECTOR OPERATIONS
//

//
// Multiplication by a vector
//

// B = A * diag(v)
template <typename T>
void DiagonalProduct(const DenseMatrix<T> &a, const DenseVector<T> &v,
                     DenseMatrix<T> *b) {
  CHECK(a.NumColumn() == v.Size());
  if ((b->NumColumn() != a.NumColumn())
      || (b->NumRow() != a.NumRow()))
    b->Reset(a.NumRow(), a.NumColumn(), 0);

  //TODO(rostami): See if lapack banded matrix multiply is faster.
  for (size_t i = 0; i < v.Size(); ++i)
    for (size_t j = 0; j < a.NumRow(); ++j)
      b->SetElement(j, i, a.Element(j, i) * v.Element(i));
}

// y = alpha * A * x + beta * y
template <typename T>
void MatrixVectorProduct(T alpha,
             const DenseMatrix<T> &a,
             const DenseVector<T> &x,
             T beta,
             DenseVector<T> *y) {
  CHECK(a.NumColumn() == x.Size());
  y->Resize(a.NumRow(), 0); // force MutateCheck
  BlasMatrixInfo<T> info_a;
  BlasVectorInfo<T> info_x, info_y;
  a.SetBlasInfo(&info_a);
  x.SetBlasInfo(&info_x);
  y->SetBlasInfo(&info_y);


  Gemv<T>()(&(info_a.transpose),
            &(info_a.leading_dimension),
            &(info_a.other_dimension),
            &alpha,
            info_a.array,
            &(info_a.leading_dimension),
            info_x.array,
            &(info_x.increment),
            &beta,
            info_y.array,
            &(info_y.increment));
}

// y = A * x
template <typename T>
void Product(const DenseMatrix<T> &a,
             const DenseVector<T> &x,
             DenseVector<T> *y) {
  MatrixVectorProduct<T>(static_cast<T>(1.0), a, x, static_cast<T>(0.0), y); 
  //Product(static_cast<T>(1.0), a, x, static_cast<T>(0.0), y);
}

// A = x * y^t
template <typename T>
void OuterProduct(const DenseVector<T> &x,
                  const DenseVector<T> &y,
                  DenseMatrix<T> *a) {
  if ((a->NumRow() != x.Size()) || (a->NumColumn() != y.Size()))
    a->Reset(x.Size(), y.Size(), 0);

  for (size_t i = 0; i < x.Size(); ++i)
    for (size_t j = 0; j < y.Size(); ++j)
      a->SetElement(i, j, x.Element(i) * y.Element(j));
}

//
//  LAPACK OPERATIONS
//

//
//  Solve linear equations, square A
//
//  A * X = B
//
template <typename T>
void SolveLinearEq(const DenseMatrix<T> &a,
                   const DenseMatrix<T> &b,
                   DenseMatrix<T> *x) {
  CHECK(a.NumRow() == b.NumRow());
  DenseMatrix<T> mut_a(a);
  DenseMatrix<T> mut_b(b);
  MutableBlasMatrixInfo<T> info_a, info_b;
  mut_a.SetBlasInfo(&info_a);
  mut_b.SetBlasInfo(&info_b);
  int32 n = mut_a.NumRow();
  int32 nrhs = mut_b.NumColumn();
  int32 info;
  char uplo = 'U';
  int32 ipiv[n];

  Gesv<T>()(&n, &nrhs, info_a.array, &(info_a.leading_dimension), ipiv,
            info_b.array, &(info_b.leading_dimension), &info);

  if(info < 0)
    LOG(ERROR) << "Error in solving linear equations:"
         << " BAD_FORMAT" << endl;
  if(info > 0)
    LOG(ERROR) << "Error in solving linear equations:"
         << " Non-PDS, increase ridge parameter" << endl;

  *x = mut_b;
}

//
//  Solve linear equations, symmetric PDS A
//
//  A * X = B
//
template <typename T>
void SolveLinearEqPDS(const DenseMatrix<T> &a,
                      const DenseMatrix<T> &b,
                      DenseMatrix<T> *x) {
  CHECK(a.NumRow() == b.NumRow());
  DenseMatrix<T> mut_a(a);
  DenseMatrix<T> mut_b(b);
  MutableBlasMatrixInfo<T> info_a, info_b;
  mut_a.SetBlasInfo(&info_a);
  mut_b.SetBlasInfo(&info_b);
  int32 n = mut_a.NumRow();
  int32 nrhs = mut_b.NumColumn();
  int32 info;
  char uplo = 'U';

  Posv<T>()(&uplo, &n, &nrhs, info_a.array, &(info_a.leading_dimension),
            info_b.array, &(info_b.leading_dimension), &info);

  if(info < 0)
    LOG(ERROR) << "Error in solving linear equations:"
         << " BAD_FORMAT" << endl;
  if(info > 0)
    LOG(ERROR) << "Error in solving linear equations:"
         << " Non-PDS, increase ridge parameter" << endl;

  *x = mut_b;
}

//  Solve linear equations, symmetric PDS A
//
//  A * x = b
//
//  Overloaded to work with vectors x and b.
template <typename T>
void SolveLinearEqPDS(const DenseMatrix<T> &A,
                      const DenseVector<T> &b,
                      DenseVector<T> *x) {

  DenseMatrix<T> B(b);
  DenseMatrix<T> X;
  SolveLinearEqPDS(A, B, &X);

  x->Clear();
  x->Resize(X.NumRow());
  for (int i = 0; i < X.NumRow(); ++i)
    x->SetElement(i, X.Element(i, 0));

}

//
// Compute eigenvalue/eigenvectors of real symmetric matrix A
//
// Evec * Eval = A * Evec
//
template <typename T>
void SymEigen(const DenseMatrix<T> &a,
                        DenseMatrix<T> *evectors,
                        DenseVector<T> *evalues) {
  CHECK(a.NumRow() == a.NumColumn());
  DenseMatrix<T> mut_a(a);
  MutableBlasMatrixInfo<T> info_a;
  mut_a.SetBlasInfo(&info_a);
  int32 n = mut_a.NumRow();
  int32 info;
  char uplo = 'U';
  char jobz = 'V';
  T *w = new T[n];
  T *work_query = new T[1];
  int32 lwork = -1;

  //first call queries for optimal work array size
  Syev<T>()(&jobz, &uplo, &n, info_a.array, &(info_a.leading_dimension),
            w, work_query, &lwork, &info);

  if(info != 0)
    LOG(ERROR) << "Error (Syev, query phase): code -- " << info << endl;

  //second call does the actual work
  T *work = new T[static_cast<size_t>(work_query[0])];
  lwork = static_cast<int32>(work_query[0]);
  //lwork = work_query[0];

  Syev<T>()(&jobz, &uplo, &n, info_a.array, &(info_a.leading_dimension),
            w, work, &lwork, &info);

  if(info != 0)
     LOG(ERROR) << "Error (Syev, work phase): code -- " << info << endl;

  *evectors = mut_a;
  evalues->Resize(n, 0);
  for (size_t i = 0; i < n; i++)
    evalues->SetElement(i, w[i]);

  delete[] w;
  delete[] work_query;
  delete[] work;
}

//
// Compute the low-rank approximation to a real symmetric matrix, given
// the sampled indices and data.
//
// Let K = [W, G12; G12^T, G22] and C = [W, G12],
// where K is n-by-n and W is l-by-l.
//
// Nystrom approximation: K_approx = C * W^+ * C^T
//
// Approximate SVD: Evec_approx = sqrt(l / n) * C * Evec(W) * Eval(W)^+
//                  Eval_approx = n / l * Eval(W)
//
// So that:  K_approx = Evec_approx * Eval_approx * Evec_approx^T
//
// On input, C contains columns sampled from the kernel matrix, and the
// vector index indicates which columns were sampled.
// On output C contains the approximate eigenvectors, and the DenseVector
// evalues holds the approximate eigenvalues.
template <typename T, typename I>
void ApproxSymEigen(const DenseVector<I> &index,
                    DenseMatrix<T> *C, DenseVector<T> *evalues) {
  CHECK(C->NumColumn() == index.Size());
  DenseMatrix<T> W(C->NumColumn(), C->NumColumn(), 0);
  DenseMatrix<T> Empty(0, 0, 0); //acts as NULL
  DenseVector<size_t> range;
  MakeRange(0, C->NumColumn(), &range);
  SubMatrixCopy(*C, index, range, &W);

  DenseMatrix<T> V;
  SymEigen(W, &V, evalues);
  W = Empty;
  DenseVector<T> evalues_inv(*evalues);
  //eigenvalues of pseudoinverse
  //TODO(rostami): decide if this is a good threshold
  //double thresh = pow(2, -23) * Max(*evalues) * W.NumRow();
  double thresh = 0;
  for (size_t i = 0; i < evalues->Size(); i++)
    if (evalues->Element(i) > thresh)
      evalues_inv.SetElement(i, 1 / evalues->Element(i));
    else
      evalues_inv.SetElement(i, 0);

  ScalarProduct(static_cast<T>(C->NumRow()) / static_cast<T>(C->NumColumn()),
                evalues);

  DenseMatrix<T> temp, evalues_inv_mat;
  Diagonal(evalues_inv, &evalues_inv_mat);

  MatrixMatrixProduct(
     static_cast<T>(
       sqrt(static_cast<T>(C->NumColumn()) / static_cast<T>(C->NumRow()))),
     V, evalues_inv_mat, static_cast<T>(0), &temp);
  V = Empty;
  evalues_inv_mat = Empty;

  DenseMatrix<T> C_t, temp_t;
  C->Transpose(&C_t);
  temp.Transpose(&temp_t);
  *C = Empty;
  temp = Empty;

  //In place product
  Product(temp_t, &C_t);
  C_t.Transpose(&temp);
  *C = temp;
}

// Compute the inverse of a general matrix, via the LU factorization.
template <typename T>
void Inverse(const DenseMatrix<T> &a, DenseMatrix<T> *a_inv) {
  CHECK(a.NumRow() == a.NumColumn());
  DenseMatrix<T> mut_a(a);
  MutableBlasMatrixInfo<T> info_a;
  mut_a.SetBlasInfo(&info_a);
  int32 n = mut_a.NumRow();
  int32 info;
  int32 *ipiv;
  ipiv = new int32[n];

  Getrf<T>()(&n, &n, info_a.array, &(info_a.leading_dimension), ipiv, &info);
  if(info != 0)
    LOG(ERROR) << "Error (Inverse, LU): code -- " << info << endl;

  int32 lwork = -1;
  T *work;
  work = new T[1];
  Getri<T>()(&n, info_a.array, &(info_a.leading_dimension), ipiv, work, &lwork,
             &info);
  if(info != 0)
    LOG(ERROR) << "Error (Inverse, Blocksize Query): code -- " << info << endl;

  lwork = static_cast<int32>(work[0]);
  //lwork = work[0];
  delete[] work;
  work = new T[lwork];
  Getri<T>()(&n, info_a.array, &(info_a.leading_dimension), ipiv, work, &lwork,
             &info);
  if(info != 0)
    LOG(ERROR)<< "Error (Inverse, Inv): code -- " << info << endl;

  delete[] ipiv;
  delete[] work;

  *a_inv = mut_a;
}

// Compute the inverse of a positive definite symmetric matrix, via
// the Cholesky factorization.
template <typename T>
void PDSInverse(const DenseMatrix<T> &a, DenseMatrix<T> *a_inv) {
  CHECK(a.NumRow() == a.NumColumn());
  DenseMatrix<T> mut_a(a);
  MutableBlasMatrixInfo<T> info_a;
  mut_a.SetBlasInfo(&info_a);
  int32 n = mut_a.NumRow();
  int32 info;
  char uplo = 'U';

  //Cholesky
  Potrf<T>()(&uplo, &n, info_a.array, &(info_a.leading_dimension), &info);
  if(info != 0)
    LOG(ERROR) << "Error (PDSInverse, Chol): code -- " << info << endl;

  //Inversion
  Potri<T>()(&uplo, &n, info_a.array, &(info_a.leading_dimension), &info);
  if(info != 0)
    LOG(ERROR) << "Error (PDSInverse, Inv): code -- " << info << endl;

  *a_inv = mut_a;

  //fill in lower triangular region
  for (size_t j = 0; j < a_inv->NumColumn() - 1; ++j)
    for (size_t i = j + 1; i < a_inv->NumRow(); ++i)
      a_inv->SetElement(i, j, a_inv->Element(j, i));
}

// Compute Cholesky factorization of PDS matrix X. That is, find upper
// triangular M such that M^T M = X.
template <typename T>
void Cholesky(const DenseMatrix<T> &X, DenseMatrix<T> *M) {
  CHECK(X.NumRow() == X.NumColumn());
  DenseMatrix<T> mut_X(X);
  MutableBlasMatrixInfo<T> info_X;
  mut_X.SetBlasInfo(&info_X);
  int32 n = mut_X.NumRow();
  int32 info;
  char uplo = 'U';

  Potrf<T>()(&uplo, &n, info_X.array, &(info_X.leading_dimension), &info);
  if(info != 0)
    LOG(ERROR) << "Error (PDSInverse, Chol): code -- " << info << endl;

  *M = mut_X;

	//set lower diagonal portion equal to zero (not cleared by default)
	for (int j = 0; j < n; ++j)
	  for (int i = j + 1; i < n; ++i)
		  M->SetElement(i, j, 0);
}

}  // namespace kernel

#endif  // KERNEL_LIB_DENSE_MATRIX_H_
