// dense-vector.h
//

// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Author: allauzen@google.com (Cyril Allauzen)
//
// \file
// A simple reference-counted dense vector representation

#ifndef KERNEL_LIB_DENSE_VECTOR_H_
#define KERNEL_LIB_DENSE_VECTOR_H_

#include "kernel/lib/blas.h"
#include "kernel/lib/kernel.h" // change to min req

namespace kernel {

//
// DENSE VECTOR CLASS
//

// Forward declaration
template <typename T>
class DenseVectorData;


// DenseVector: a simple reference-counted dense vector representation
template <typename T>
class DenseVector {
 public:
  typedef T ElementType;
  typedef size_t IndexType;

  DenseVector() : data_(new DenseVectorData<T>()) {}

  DenseVector(size_t n, const T& x = T()) : data_(new DenseVectorData<T>()) {
    data_->Resize(n, x);
  }

  DenseVector(const DenseVector<T> &v) : data_(v.data_) {
    if (data_)
      data_->IncrRefCount();
  }

  ~DenseVector() {
    if (!data_->DecrRefCount())
      delete data_;
  }

  DenseVector<T> &operator=(const DenseVector<T> &v) {
    if (this != &v) {
      if (!data_->DecrRefCount()) delete data_;
      v.data_->IncrRefCount();
      data_ = v.data_;
    }
    return *this;
  }

  size_t Size() const { return data_->Size(); }

  void Clear() {
    MutateCheck();
    data_->Clear();
  }

  void Resize(size_t n, const T &x = 0.0F) {
    MutateCheck();
    data_->Resize(n, x);
  }

  const T &Element(size_t i) const {
    CHECK(i < data_->Size());
    return data_->Element(i);
  }

  void SetElement(size_t i, const T &x) {
    CHECK(i < data_->Size());
    MutateCheck();
    data_->SetElement(i, x);
  }

  void AddElement(size_t i, const T &x) {
    MutateCheck();
    if (i >= Size())
      data_->Resize(i + 1, static_cast<T>(0.0F));
    data_->SetElement(i, x);
  }

  void SetBlasInfo(BlasVectorInfo<T> *info) const {
    info->dimension = Size();
    info->increment = 1;
    info->array = data_->CArray();
  }

  static DenseVector<T> *Read(istream &strm) {
    return new DenseVector<T>(DenseVectorData<T>::Read(strm));
  }

  bool Write(ostream &strm) const {
    return data_->Write(strm);
  }

 private:
  DenseVector(DenseVectorData<T> *data) : data_(data) {}

  void MutateCheck() {
    // Copy on write
    if (data_->RefCount() > 1) {
      data_->DecrRefCount();
      data_ = new DenseVectorData<T>(*data_);
    }
  }

  DenseVectorData<T> *data_;
};


// Provides the data representation of a dense vector as a STL vector
template <typename T>
class DenseVectorData {
 public:
  DenseVectorData() {}

  explicit DenseVectorData(const DenseVectorData<T> &data)
      : elements_(data.elements_) {}

  ~DenseVectorData() {}

  size_t Size() const { return elements_.size(); }
  void Reserve(size_t n) { elements_.reserve(n); }
  void Resize(size_t n, const T &x) { elements_.resize(n, x); }
  void Clear() { elements_.clear(); }

  const T &Element(size_t i) const { return elements_[i]; }
  void SetElement(size_t i, const T &x) { elements_[i] = x; }

  int RefCount() const { return ref_count_.count(); }
  int IncrRefCount() { return ref_count_.Incr(); }
  int DecrRefCount() { return ref_count_.Decr(); }

  T *CArray() const { return const_cast<T*>(&(elements_[0])); }

  static DenseVectorData<T> *Read(istream &strm) {
    DenseVectorData<T> *data = new DenseVectorData<T>;
    fst::ReadType(strm, &(data->elements_));
    if (!strm) {
      LOG(ERROR) << "DenseVector::Read: Read failed!";
      return 0;
    }
    return data;
  }

  bool Write(ostream &strm) const {
    fst::WriteType(strm, elements_);
    if (!strm) {
      LOG(ERROR) << "DenseVector::Write: write failed!";
      return false;
    }
    return true;
  }

 private:
  vector<T> elements_;
  RefCounter ref_count_;

  void operator=(const DenseVectorData<T> &);  // disallow
};


//
// VECTOR OPERATIONS
//

// Fill vector with integers (i, i + 1, ..., j).
template <typename T> inline
void MakeRange(size_t i, size_t j, DenseVector<T> *v) {
  CHECK(j > i);
  v->Resize(j - i, 0);
  for (size_t k = 0; k < j - i; ++k)
    v->SetElement(k, i + k);
}

// If j > size, then fill vector with integers
// (0, ..., j - size, i, ..., size)
template <typename T> inline
void MakeRangeWrap(size_t i, size_t j, size_t size, DenseVector<T> *v) {
  CHECK(j > i);
  CHECK(i <= size);
  v->Clear();
  size_t vi = 0;
  if (j >= size) {
    CHECK(j - size < size); //check does not wrap around twice
    for (size_t k = 0; k < j - size; ++k) {
      v->AddElement(vi, k);
      ++vi;
    }
  }
  for (size_t k = i; k < min(size, j); ++k) {
    v->AddElement(vi, k);
    ++vi;
  }
}

// Copy a subset of a vector.
template <typename T> inline
void SubsetCopy(size_t i, size_t j,
                const DenseVector<T> &source,
                DenseVector<T> *subset) {
  CHECK(j > i);
  CHECK(i >= 0);
  CHECK(j < source.Size());
  if (subset->Size() != j - i)
    subset->Resize(j - i, 0);
  for (size_t k = i; k < j; ++k)
    subset->SetElement(k - i, source.Element(k));
}

// Copy a subset of a vector.
template <typename T, typename I> inline
void SubsetCopy(const DenseVector<T> &source,
                const DenseVector<I> index,
                DenseVector<T> *subset) {
  subset->Resize(index.Size(), 0);
  for (size_t i = 0; i < index.Size(); ++i)
    subset->SetElement(i, source.Element(index.Element(i)));
}

// Subtract the mean of v from element of v.
template <typename T> inline
void Center(DenseVector<T> *v) {
  T mean = 0;
  for (size_t i = 0; i < v->Size(); ++i)
    mean += v->Element(i);

  mean = mean / v->Size();

  for (size_t i = 0; i < v->Size(); ++i)
    v->SetElement(i, v->Element(i) - mean);
}


// Vector equality
template <typename T> inline
bool Equal(const DenseVector<T> &a, const DenseVector<T> &b) {
  if (a.Size() != b.Size()) return false;
  for (size_t i = 0; i < a.Size(); ++i)
    if (a.Element(i) != b.Element(i)) return false;
  return true;
}

// Vector approximate equality
template <typename T> inline
bool ApproxEqual(const DenseVector<T> &a, const DenseVector<T> &b, T delta) {
  if (a.Size() != b.Size())
    return false;
  for (size_t i = 0; i < a.Size(); ++i)
    if ((a.Element(i) > b.Element(i) + delta) ||
        (a.Element(i) < b.Element(i) - delta)) {
      VLOG(1) << "Element " << i << " differs: ("
           << a.Element(i) << "," << b.Element(i) << ")" << endl;
      return false;
    }
  return true;
}

// Return largest element in vector, linear time
template <typename T> inline
T Max(const DenseVector<T> &v) {
  T max = 0;
  for (size_t i = 0; i < v.Size(); i++)
    if (max < v.Element(i))
      max = v.Element(i);
  return max;
}

// Return smallest element in vector, linear time
template <typename T> inline
T Min(const DenseVector<T> &v) {
  T min = 0;
  for (size_t i = 0; i < v.Size(); i++)
    if (min > v.Element(i))
      min = v.Element(i);
  return min;
}

// Return the l2 norm
template <typename T> inline
T Norm(const DenseVector<T> &v) {
  T norm = 0;
  for (size_t i = 0; i < v.Size(); ++i)
    norm += pow(v.Element(i), 2);
  return sqrt(norm);
}

//
// Multiplication by a scalar
//

// y = alpha * x
template <typename T> inline
void ScalarProduct(const T &alpha,
                   const DenseVector<T> &x,
                   DenseVector<T> *y) {
  if(y->Size() != x.Size())
    y->Resize(x.Size(), 0);

  for (size_t i = 0; i < x.Size(); ++i)
    y->SetElement(i, alpha * x.Element(i));
}

// x = alpha * x
template <typename T> inline
void ScalarProduct(const T &alpha, DenseVector<T> *x) {
  ScalarProduct(alpha, *x, x);
}


//
// Addition
//

// z = alpha * x + y
template <typename T> inline
void Sum(const T &alpha,
         const DenseVector<T> &x,
         const DenseVector<T> &y,
         DenseVector<T> *z){
  size_t m = x.Size() < y.Size() ? x.Size() : y.Size();
  size_t n = x.Size() > y.Size() ? x.Size() : y.Size();
  if (z->Size() < n)
    z->Resize(n, 0.0F);
  for (size_t i = 0; i < m; ++i)
    z->SetElement(i, alpha * x.Element(i) + y.Element(i));
  for (size_t i = m; i < x.Size(); ++i)
    z->SetElement(i, alpha * x.Element(i));
  for (size_t i = m; i < y.Size(); ++i)
    z->SetElement(i, y.Element(i));
}

// y = alpha * x + y
template <typename T> inline
void Sum(const T &alpha, const DenseVector<T> &x, DenseVector<T> *y) {
  Sum(alpha, x, *y, y);
}

// z = x + y
template <typename T> inline
void Sum(const DenseVector<T> &x, const DenseVector<T> &y, DenseVector<T> *z) {
  Sum<T>(1.0F, x, y, z);
}

// y = x + y
template <typename T> inline
void Sum(const DenseVector<T> &x, DenseVector<T> *y) {
  Sum<T>(1.0F, x, *y, y);
}


//
// Dot product
//

// dot = x^T * y
template <typename T> inline
T DotProduct(const DenseVector<T> &x, const DenseVector<T> &y) {
  T dot = 0.0F;
  size_t n = x.Size() <= y.Size() ? x.Size() : y.Size();
  for (size_t i = 0; i < n; ++i)
    dot += x.Element(i) * y.Element(i);
  return dot;
}

// dot = x^T * Diag(y) * z
template <typename T> inline
T DotProduct(const DenseVector<T> &x,
             const DenseVector<T> &y,
             const DenseVector<T> &z) {
  T dot = 0.0F;
  size_t n =
      x.Size() <= y.Size()
      ? (z.Size() <= x.Size() ? z.Size() :x.Size())
      : (z.Size() <= y.Size() ? z.Size() : y.Size());
  for (size_t i = 0; i < n; ++i)
    dot += x.Element(i) * y.Element(i) * z.Element(i);
  return dot;
}

// z = x .* y,   pointwise product of x and y.
template <typename T> inline
void PointwiseProduct(const DenseVector<T> &x,
                      const DenseVector<T> &y,
                      DenseVector<T> *z) {
  CHECK(x.Size() == y.Size());
  z->Resize(x.Size(), 0);
  for (size_t i = 0; i < z->Size(); ++i)
    z->SetElement(i, x.Element(i) * y.Element(i));
}

// Sum of entries
template <typename T> inline
T Sum(const DenseVector<T> &v) {
  T sum = 0;
  for (size_t i = 0; i < v.Size(); ++i)
    sum += v.Element(i);
  return sum;
}

// Return +/-1 vector w, indicating the signs of entries of v.
template <typename T> inline
void SignVector(const DenseVector<T> &v, DenseVector<T> *w) {
  w->Resize(v.Size(), 0);
  for (size_t i = 0; i < v.Size(); ++i) {
    if (v.Element(i) > 0)
      w->SetElement(i, 1);
    else
      w->SetElement(i, -1);
  }
}

// Mean of entries.
template <typename T> inline
T Mean(const DenseVector<T> &v) {
  T mean = Sum(v);
  mean = mean / v.Size();
  return mean;
}

// Standard deviation of entries.
template <typename T> inline
T StdDev(const DenseVector<T> &v) {
  T std;
  T mean = Mean(v);
  T ss = 0;
  for (size_t i = 0; i < v.Size(); ++i)
    ss += pow(v.Element(i), 2);

  ss = ss / v.Size();
  std = sqrt(ss - pow(mean, 2));
  return std;
}

}  // namespace kernel

#endif  // KERNEL_LIB_DENSE_VECTOR_H_
