// properties.cc
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Author: allauzen@cs.nyu.edu (Cyril Allauzen)
//
// \file
// Functions definiting the properties for kernel operations.

#include "kernel/lib/properties.h"

namespace kernel {

uint64 ExternalProductProperties(uint64 inprops, double a) {
  uint64 outprops = inprops & (kSymmetric | kAsymmetric);

  if (a >= 0.0F) {
    outprops |= inprops & (kPosDefinite | kNegDefinite | kNotDefinite);
  } else if (inprops & kPosDefinite) {
    outprops |= kNegDefinite;
  } else
    outprops |= (inprops & kNotDefinite);

  if (a == 1.0F)
    outprops |= outprops & (kNormalized | kUnnormalized);
  else if (inprops & kNormalized)
    outprops |= kUnnormalized;

  return outprops;
}

uint64 ProductProperties(uint64 inprops1, uint64 inprops2) {
  uint64 outprops = 0;
  if ((inprops1 & kSymmetric) && (inprops2 & kSymmetric))
    outprops |= kSymmetric;
  else if ((inprops1 & kAsymmetric) && (inprops2 & kSymmetric))
    outprops |= kAsymmetric;
  else if ((inprops1 & kSymmetric) && (inprops2 & kAsymmetric))
    outprops |= kAsymmetric;

  if ((inprops1 & kNormalized) && (inprops2 & kNormalized))
    outprops |= kNormalized;
  else if ((inprops1 & kUnnormalized) && (inprops2 & kNormalized))
    outprops |= kUnnormalized;
  else if ((inprops1 & kNormalized) && (inprops2 & kUnnormalized))
    outprops |= kUnnormalized;

  if ((inprops1 & kPosDefinite) && (inprops2 & kPosDefinite))
    outprops |= kPosDefinite;

  return outprops;
}

uint64 SumProperties(uint64 inprops1, uint64 inprops2, double a1, double a2) {
  if (a1 == 0.0F)
    return ExternalProductProperties(inprops2, a2);
  if (a2 == 0.0F)
    return ExternalProductProperties(inprops1, a1);

  uint64 outprops = 0;
  if ((inprops1 & kSymmetric) && (inprops2 & kSymmetric))
    outprops |= kSymmetric;
  else if ((inprops1 & kAsymmetric) && (inprops2 & kSymmetric))
    outprops |= kAsymmetric;
  else if ((inprops1 & kSymmetric) && (inprops2 & kAsymmetric))
    outprops |= kAsymmetric;

  if ((a1 >= 0.0F) && (a2 >= 0.0F) &&
      (inprops1 & kNormalized) &&
      (inprops2 & kNormalized))
    outprops |= (a1 + a2 == 1.0F)
                ? kNormalized
                : kUnnormalized;
  if ((a1 <= 0.0F) && (a2 <= 0.0F) &&
      (inprops1 & kNormalized) &&
      (inprops2 & kNormalized))
    outprops |= kUnnormalized;

  uint64 props1 = ExternalProductProperties(inprops1, a1);
  uint64 props2 = ExternalProductProperties(inprops2, a2);
  if ((props1 & kPosDefinite) && (props2 & kPosDefinite))
    outprops |= kPosDefinite;
  else if ((props1 & kNegDefinite) && (props2 & kNegDefinite))
    outprops |= kNegDefinite;

  return outprops;
}

}  // namespace kernel
